import java.util.Comparator;
import java.util.List;
import java.util.Objects;

import static java.util.Comparator.comparingInt;

public class Point /* implements Comparable<Point> */ {
    public double x, y;

    Point() {
        this(0, 0);
    }

    Point(double x, double y) {
        this.x = x;
        this.y = y;
    }

    public Point add(Point other) {
        return new Point(this.x + other.x, this.y + other.y);
    }

    public Point subtract(Point other) {
        return new Point(this.x - other.x, this.y - other.y);
    }

    public double distanceTo(Point other) {
        return Math.sqrt((this.x - other.x)*(this.x - other.x) + (this.y - other.y)*(this.y - other.y));
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        Point point = (Point) o;
        return Double.compare(point.x, x) == 0 &&
                Double.compare(point.y, y) == 0;
    }

    @Override
    public int hashCode() {
        return Objects.hash(x, y);
    }

//    @Override
//    public int compareTo(Point o) {
//        return Comparator.<Point>comparingDouble(p -> p.x).thenComparingDouble(p -> p.y).compare(this, o);
//    }


    @Override
    public String toString() {
        return "(" + x + ", " + y + ")";
    }

    public static Point intersect(Point A, Point B, Point C, Point D) {
        return segmentIntersection(A, B, C, D);
    }

    /**
     * Determine the point of intersection of two segments AB and CD
     *
     * Solution with inspiration from:
     * https://stackoverflow.com/a/1968345/13895241
     *
     * @param a point A of AB
     * @param b point B of AB
     * @param c point C of CD
     * @param d point D of CD
     * @return the point of intersection or null
     */
    public static Point segmentIntersection(Point a, Point b, Point c, Point d) {
        Point s1 = b.subtract(a);
        Point s2 = d.subtract(c);

        double s, t;
        s = (-s1.y * (a.x - c.x) + s1.x * (a.y - c.y)) / (-s2.x * s1.y + s1.x * s2.y);
        t = ( s2.x * (a.y - c.y) - s2.y * (a.x - c.x)) / (-s2.x * s1.y + s1.x * s2.y);

        if (s >= 0 && s <= 1 && t >= 0 && t <= 1) {
            return new Point(a.x + (t * s1.x), a.y + (t * s1.y));
        }

        return null;
    }

    private static double area(Point p1, Point p2, Point p3) {
        double a = p1.distanceTo(p2);
        double b = p2.distanceTo(p3);
        double c = p1.distanceTo(p3);

        double s = (a + b + c) / 2;

        return Math.sqrt(s * (s - a) * (s - b) * (s - c));
    }

    public static double area(List<Point> p) {
        var basePoint = p.get(0);

        double result = 0;

        for (int i = 1; i < p.size() - 1; i++) {
            result += area(basePoint, p.get(i), p.get(i + 1));
        }

        return result;
    }

    public static double perimeter(List<Point> p) {
        double result = 0;

        for (int i = 0; i < p.size() - 1; i++) {
            result += p.get(i).distanceTo(p.get(i + 1));
        }
        result += p.get(0).distanceTo(p.get(p.size() - 1));

        return result;
    }
}
